from odict import odict
from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.controller import Controller
from yafowil.tests import YafowilTestCase
from yafowil.tests import fxml
from yafowil.utils import Tag
import yafowil.common
import yafowil.compound  # noqa


###############################################################################
# Helpers
###############################################################################

tag = Tag(lambda msg: msg)


###############################################################################
# Tests
###############################################################################

class TestCompound(YafowilTestCase):

    def test_compound_blueprint_value_via_compound(self):
        # Render Compound with values set via compound widget
        value = {
            'inner': 'Value 1 from parent',
            'inner2': 'Value 2 from parent',
        }
        compound = factory(
            'compound',
            name='COMPOUND',
            value=value)
        compound['inner'] = factory('text')
        compound['inner2'] = factory(
            'text',
            props={
                'required': True
            })
        self.checkOutput("""
        <div>
          <input class="text" id="input-COMPOUND-inner" name="COMPOUND.inner"
                 type="text" value="Value 1 from parent"/>
          <input class="required text" id="input-COMPOUND-inner2"
                 name="COMPOUND.inner2" required="required" type="text"
                 value="Value 2 from parent"/>
        </div>
        """, fxml(tag('div', compound())))

    def test_compound_blueprint_value_via_members(self):
        # Render Compound with values set via compound members
        compound = factory(
            'compound',
            name='COMPOUND')
        compound['inner'] = factory(
            'text',
            value='value1')
        compound['inner2'] = factory(
            'error:text',
            value='value2',
            props={
                'required': True
            })
        self.checkOutput("""
        <div>
          <input class="text" id="input-COMPOUND-inner" name="COMPOUND.inner"
                 type="text" value="value1"/>
          <input class="required text" id="input-COMPOUND-inner2"
                 name="COMPOUND.inner2" required="required" type="text"
                 value="value2"/>
        </div>
        """, fxml(tag('div', compound())))

    def test_compound_blueprint_value_conflict(self):
        # ValueError if value for a compound member is defined both
        value = {'inner': 'Value 1 from parent'}
        compound = factory(
            'compound',
            name='COMPOUND',
            value=value)
        compound['inner'] = factory(
            'text',
            value='value1')
        with self.assertRaises(ValueError) as arc:
            compound()
        msg = "Both compound and compound member provide a value for 'inner'"
        self.assertEqual(str(arc.exception), msg)

    def test_compound_blueprint_extraction(self):
        compound = factory('compound', name='COMPOUND')
        compound['inner'] = factory('text', value='value1')
        compound['inner2'] = factory(
            'error:text',
            value='value2',
            props={
                'required': True
            })

        # Extract Compound with empty request
        data = compound.extract({})
        self.assertEqual(data.name, 'COMPOUND')
        self.assertEqual(data.value, UNSET)
        expected = odict()
        expected['inner'] = UNSET
        expected['inner2'] = UNSET
        self.assertEqual(data.extracted, expected)
        self.assertEqual(data.errors, [])

        inner_data = data['inner']
        self.assertEqual(inner_data.name, 'inner')
        self.assertEqual(inner_data.value, 'value1')
        self.assertEqual(inner_data.extracted, UNSET)
        self.assertEqual(inner_data.errors, [])

        # Extract with a value in request
        request = {
            'COMPOUND.inner': 'newvalue',
            'COMPOUND.inner2': '',
        }
        data = compound.extract(request)
        data_inner = data['inner']
        self.assertEqual(data_inner.name, 'inner')
        self.assertEqual(data_inner.value, 'value1')
        self.assertEqual(data_inner.extracted, 'newvalue')
        self.assertEqual(data_inner.errors, [])

        data_inner2 = data['inner2']
        self.assertEqual(data_inner2.name, 'inner2')
        self.assertEqual(data_inner2.value, 'value2')
        self.assertEqual(data_inner2.extracted, '')
        self.assertEqual(
            data_inner2.errors,
            [ExtractionError('Mandatory field was empty')]
        )

        expected = odict()
        expected['inner'] = 'newvalue'
        expected['inner2'] = ''
        self.assertEqual(data.extracted, expected)

        self.checkOutput("""
        <div>
          <input class="text" id="input-COMPOUND-inner" name="COMPOUND.inner"
                 type="text" value="newvalue"/>
          <div class="error">
            <div class="errormessage">Mandatory field was empty</div>
            <input class="required text" id="input-COMPOUND-inner2"
                   name="COMPOUND.inner2" required="required"
                   type="text" value=""/>
          </div>
        </div>
        """, fxml('<div>' + compound(data=data) + '</div>'))

    def test_compound_blueprint_display_rendering(self):
        # Compound display renderers, same as edit renderers
        compound = factory(
            'compound',
            name='COMPOUND',
            mode='display')
        self.assertEqual(tag('div', compound()), '<div></div>')

    def test_compound_blueprint_structural_children(self):
        # Compound with structural compound as child
        value = {
            'inner': 'Value 1 from parent',
            'inner2': 'Value 2 from parent',
        }
        compound = factory(
            'compound',
            name='COMPOUND',
            value=value)
        structural = compound['STRUCTURAL'] = factory(
            'compound',
            props={
                'structural': True
            })
        structural['inner'] = factory('text')
        structural['inner2'] = factory(
            'text',
            props={
                'required': True
            })
        self.checkOutput("""
        <div>
          <input class="text" id="input-COMPOUND-inner" name="COMPOUND.inner"
                 type="text" value="Value 1 from parent"/>
          <input class="required text" id="input-COMPOUND-inner2"
                 name="COMPOUND.inner2" required="required" type="text"
                 value="Value 2 from parent"/>
        </div>
        """, fxml(tag('div', compound())))

        self.assertEqual(compound.treerepr().split('\n'), [
            "<class 'yafowil.base.Widget'>: COMPOUND",
            "  <class 'yafowil.base.Widget'>: STRUCTURAL",
            "    <class 'yafowil.base.Widget'>: inner",
            "    <class 'yafowil.base.Widget'>: inner2",
            ""
        ])

        data = compound.extract({
            'COMPOUND.inner': 'newvalue',
            'COMPOUND.inner2': '',
        })
        self.assertEqual(data.name, 'COMPOUND')
        self.assertEqual(data.value, {
            'inner2': 'Value 2 from parent',
            'inner': 'Value 1 from parent'
        })
        expected = odict()
        expected['inner'] = 'newvalue'
        expected['inner2'] = ''
        self.assertEqual(data.extracted, expected)

        data_inner = data['inner']
        self.assertEqual(data_inner.name, 'inner')
        self.assertEqual(data_inner.value, 'Value 1 from parent')
        self.assertEqual(data_inner.extracted, 'newvalue')
        self.assertEqual(data_inner.errors, [])

        data_inner2 = data['inner2']
        self.assertEqual(data_inner2.name, 'inner2')
        self.assertEqual(data_inner2.value, 'Value 2 from parent')
        self.assertEqual(data_inner2.extracted, '')
        self.assertEqual(
            data_inner2.errors,
            [ExtractionError('Mandatory field was empty')]
        )

    def test_compound_blueprint_compound_children(self):
        # Compound with compound as child
        value = {
            'CHILD_COMPOUND': {
                'inner': 'Value 1 from parent',
                'inner2': 'Value 2 from parent',
            }
        }
        compound = factory(
            'compound',
            name='COMPOUND',
            value=value)
        child_compound = compound['CHILD_COMPOUND'] = factory('compound')
        child_compound['inner'] = factory('text')
        child_compound['inner2'] = factory(
            'text',
            props={
                'required': True
            })
        self.checkOutput("""
        <div>
          <input class="text" id="input-COMPOUND-CHILD_COMPOUND-inner"
                 name="COMPOUND.CHILD_COMPOUND.inner" type="text"
                 value="Value 1 from parent"/>
          <input class="required text" id="input-COMPOUND-CHILD_COMPOUND-inner2"
                 name="COMPOUND.CHILD_COMPOUND.inner2" required="required"
                 type="text" value="Value 2 from parent"/>
        </div>
        """, fxml(tag('div', compound())))  # noqa

        self.assertEqual(compound.treerepr().split('\n'), [
            "<class 'yafowil.base.Widget'>: COMPOUND",
            "  <class 'yafowil.base.Widget'>: CHILD_COMPOUND",
            "    <class 'yafowil.base.Widget'>: inner",
            "    <class 'yafowil.base.Widget'>: inner2",
            ""
        ])

        data = compound.extract({
            'COMPOUND.CHILD_COMPOUND.inner': 'newvalue',
            'COMPOUND.CHILD_COMPOUND.inner2': 'newvalue2',
        })
        self.assertEqual(data.name, 'COMPOUND')
        self.assertEqual(data.value, {
            'CHILD_COMPOUND': {
                'inner2': 'Value 2 from parent',
                'inner': 'Value 1 from parent'
            }
        })
        expected = odict()
        expected['CHILD_COMPOUND'] = odict()
        expected['CHILD_COMPOUND']['inner'] = 'newvalue'
        expected['CHILD_COMPOUND']['inner2'] = 'newvalue2'
        self.assertEqual(data.extracted, expected)
        self.assertEqual(data.errors, [])

        data_compound = data['CHILD_COMPOUND']
        self.assertEqual(data_compound.name, 'CHILD_COMPOUND')
        self.assertEqual(data_compound.value, {
            'inner2': 'Value 2 from parent',
            'inner': 'Value 1 from parent'
        })
        expected = odict()
        expected['inner'] = 'newvalue'
        expected['inner2'] = 'newvalue2'
        self.assertEqual(data_compound.extracted, expected)
        self.assertEqual(data_compound.errors, [])

        data_inner = data['CHILD_COMPOUND']['inner']
        self.assertEqual(data_inner.name, 'inner')
        self.assertEqual(data_inner.value, 'Value 1 from parent')
        self.assertEqual(data_inner.extracted, 'newvalue')
        self.assertEqual(data_inner.errors, [])

        data_inner2 = data['CHILD_COMPOUND']['inner2']
        self.assertEqual(data_inner2.name, 'inner2')
        self.assertEqual(data_inner2.value, 'Value 2 from parent')
        self.assertEqual(data_inner2.extracted, 'newvalue2')
        self.assertEqual(data_inner2.errors, [])

    def test_compound_blueprint_structural_and_compound_children(self):
        # Compound with structural compound with compound as children
        value = {
            'CHILD_COMPOUND': {
                'inner': 'Value 1 from parent',
                'inner2': 'Value 2 from parent',
            }
        }
        compound = factory(
            'compound',
            name='COMPOUND',
            value=value)
        structural = compound['STRUCTURAL'] = factory(
            'compound',
            props={
                'structural': True
            })
        child_compound = structural['CHILD_COMPOUND'] = factory('compound')
        child_compound['inner'] = factory('text')
        child_compound['inner2'] = factory(
            'text',
            props={
                'required': True
            })
        self.checkOutput("""
        <div>
          <input class="text" id="input-COMPOUND-CHILD_COMPOUND-inner"
                 name="COMPOUND.CHILD_COMPOUND.inner" type="text"
                 value="Value 1 from parent"/>
          <input class="required text" id="input-COMPOUND-CHILD_COMPOUND-inner2"
                 name="COMPOUND.CHILD_COMPOUND.inner2" required="required"
                 type="text" value="Value 2 from parent"/>
        </div>
        """, fxml(tag('div', compound())))  # noqa

        self.assertEqual(compound.treerepr().split('\n'), [
            "<class 'yafowil.base.Widget'>: COMPOUND",
            "  <class 'yafowil.base.Widget'>: STRUCTURAL",
            "    <class 'yafowil.base.Widget'>: CHILD_COMPOUND",
            "      <class 'yafowil.base.Widget'>: inner",
            "      <class 'yafowil.base.Widget'>: inner2",
            ""
        ])
        self.assertEqual(
            compound['STRUCTURAL'].attrs.storage,
            {'structural': True}
        )
        self.assertEqual(
            compound['STRUCTURAL']['CHILD_COMPOUND'].attrs.storage,
            {}
        )

        data = compound.extract({
            'COMPOUND.CHILD_COMPOUND.inner': 'newvalue',
            'COMPOUND.CHILD_COMPOUND.inner2': 'newvalue2',
        })
        self.assertEqual(data.name, 'COMPOUND')
        self.assertEqual(data.value, {
            'CHILD_COMPOUND': {
                'inner2': 'Value 2 from parent',
                'inner': 'Value 1 from parent'
            }
        })
        expected = odict()
        expected['CHILD_COMPOUND'] = odict()
        expected['CHILD_COMPOUND']['inner'] = 'newvalue'
        expected['CHILD_COMPOUND']['inner2'] = 'newvalue2'
        self.assertEqual(data.extracted, expected)
        self.assertEqual(data.errors, [])

        data_compound = data['CHILD_COMPOUND']
        self.assertEqual(data_compound.name, 'CHILD_COMPOUND')
        self.assertEqual(data_compound.value, {
            'inner2': 'Value 2 from parent',
            'inner': 'Value 1 from parent'
        })
        expected = odict()
        expected['inner'] = 'newvalue'
        expected['inner2'] = 'newvalue2'
        self.assertEqual(data_compound.extracted, expected)
        self.assertEqual(data_compound.errors, [])

        data_inner = data['CHILD_COMPOUND']['inner']
        self.assertEqual(data_inner.name, 'inner')
        self.assertEqual(data_inner.value, 'Value 1 from parent')
        self.assertEqual(data_inner.extracted, 'newvalue')
        self.assertEqual(data_inner.errors, [])

        data_inner2 = data['CHILD_COMPOUND']['inner2']
        self.assertEqual(data_inner2.name, 'inner2')
        self.assertEqual(data_inner2.value, 'Value 2 from parent')
        self.assertEqual(data_inner2.extracted, 'newvalue2')
        self.assertEqual(data_inner2.errors, [])

    def test_compound_blueprint_address_compound_value_parent(self):
        # Address different compounds with value on parent
        value = {
            'c1': {
                'f1': 'Foo',
            },
            'c2': {
                'f2': 'Bar',
                'f3': 'Baz',
            },
        }
        compound = factory(
            'compound',
            'comp',
            value=value)
        compound['c1'] = factory('compound')
        compound['c1']['f1'] = factory('text')
        compound['c2'] = factory('compound')
        compound['c2']['f2'] = factory('text')
        compound['c2']['f3'] = factory('text')
        compound['c3'] = factory('compound')
        compound['c3']['f4'] = factory('text')

        self.checkOutput("""
        <div>
          <input class="text" id="input-comp-c1-f1" name="comp.c1.f1"
                 type="text" value="Foo"/>
          <input class="text" id="input-comp-c2-f2" name="comp.c2.f2"
                 type="text" value="Bar"/>
          <input class="text" id="input-comp-c2-f3" name="comp.c2.f3"
                 type="text" value="Baz"/>
          <input class="text" id="input-comp-c3-f4" name="comp.c3.f4"
                 type="text" value=""/>
        </div>
        """, fxml(tag('div', compound())))

        self.assertEqual(compound.treerepr().split('\n'), [
            "<class 'yafowil.base.Widget'>: comp",
            "  <class 'yafowil.base.Widget'>: c1",
            "    <class 'yafowil.base.Widget'>: f1",
            "  <class 'yafowil.base.Widget'>: c2",
            "    <class 'yafowil.base.Widget'>: f2",
            "    <class 'yafowil.base.Widget'>: f3",
            "  <class 'yafowil.base.Widget'>: c3",
            "    <class 'yafowil.base.Widget'>: f4",
            ""
        ])

        data = compound.extract({
            'comp.c1.f1': 'Foo 1',
            'comp.c2.f2': 'Bar 2',
            'comp.c2.f3': 'Baz 1',
        })
        self.assertEqual(data.name, 'comp')
        self.assertEqual(data.value, {
            'c2': {
                'f2': 'Bar',
                'f3': 'Baz'
            },
            'c1': {
                'f1': 'Foo'
            }
        })
        expected = odict()
        expected['c1'] = odict()
        expected['c1']['f1'] = 'Foo 1'
        expected['c2'] = odict()
        expected['c2']['f2'] = 'Bar 2'
        expected['c2']['f3'] = 'Baz 1'
        expected['c3'] = odict()
        expected['c3']['f4'] = UNSET
        self.assertEqual(data.extracted, expected)
        self.assertEqual(data.errors, [])

        # c1
        data_c1 = data['c1']
        self.assertEqual(data_c1.name, 'c1')
        self.assertEqual(data_c1.value, {'f1': 'Foo'})
        expected = odict()
        expected['f1'] = 'Foo 1'
        self.assertEqual(data_c1.extracted, expected)
        self.assertEqual(data_c1.errors, [])

        data_f1 = data['c1']['f1']
        self.assertEqual(data_f1.name, 'f1')
        self.assertEqual(data_f1.value, 'Foo')
        self.assertEqual(data_f1.extracted, 'Foo 1')
        self.assertEqual(data_f1.errors, [])

        # c2
        data_c2 = data['c2']
        self.assertEqual(data_c2.name, 'c2')
        self.assertEqual(data_c2.value, {
            'f2': 'Bar',
            'f3': 'Baz'
        })
        expected = odict()
        expected['f2'] = 'Bar 2'
        expected['f3'] = 'Baz 1'
        self.assertEqual(data_c2.extracted, expected)
        self.assertEqual(data_c2.errors, [])

        data_f2 = data['c2']['f2']
        self.assertEqual(data_f2.name, 'f2')
        self.assertEqual(data_f2.value, 'Bar')
        self.assertEqual(data_f2.extracted, 'Bar 2')
        self.assertEqual(data_f2.errors, [])

        data_f3 = data['c2']['f3']
        self.assertEqual(data_f3.name, 'f3')
        self.assertEqual(data_f3.value, 'Baz')
        self.assertEqual(data_f3.extracted, 'Baz 1')
        self.assertEqual(data_f3.errors, [])

        # c3
        data_c3 = data['c3']
        self.assertEqual(data_c3.name, 'c3')
        self.assertEqual(data_c3.value, UNSET)
        expected = odict()
        expected['f4'] = UNSET
        self.assertEqual(data_c3.extracted, expected)
        self.assertEqual(data_c3.errors, [])

        data_f4 = data['c3']['f4']
        self.assertEqual(data_f4.name, 'f4')
        self.assertEqual(data_f4.value, UNSET)
        self.assertEqual(data_f4.extracted, UNSET)
        self.assertEqual(data_f4.errors, [])

    def test_compound_blueprint_value_callbacks(self):
        # Check compound with value callbacks
        def val(widget, data):
            return 'val F1'
        value = {
            'f1': val,
        }
        compound = factory(
            'compound',
            'comp',
            value=value)
        compound['f1'] = factory('text')
        self.assertEqual(compound(), (
            '<input class="text" id="input-comp-f1" name="comp.f1" '
            'type="text" value="val F1" />'
        ))

        data = compound.extract({'comp.f1': 'New val 1'})
        self.assertEqual(data.name, 'comp')
        self.assertEqual(data.value, {'f1': val})
        expected = odict()
        expected['f1'] = 'New val 1'
        self.assertEqual(data.extracted, expected)
        self.assertEqual(data.errors, [])

        def value(widget, data):
            return {
                'f1': 'F1 Val'
            }
        compound = factory(
            'compound',
            'comp',
            value=value)
        compound['f1'] = factory('text')
        self.assertEqual(compound(), (
            '<input class="text" id="input-comp-f1" name="comp.f1" '
            'type="text" value="F1 Val" />'
        ))

        data = compound.extract({'comp.f1': 'New val 1'})
        self.assertEqual(data.name, 'comp')
        self.assertEqual(data.value, {'f1': 'F1 Val'})
        expected = odict()
        expected['f1'] = 'New val 1'
        self.assertEqual(data.extracted, expected)
        self.assertEqual(data.errors, [])

    def test_div_blueprint_compound(self):
        # Div blueprint as compound
        div = factory(
            'div',
            name='DIV_COMPOUND')
        div['inner'] = factory(
            'text',
            value='value1')
        div['inner2'] = factory(
            'text',
            value='value2',
            props={
                'required': True
            })
        self.checkOutput("""
        <div>
          <input class="text" id="input-DIV_COMPOUND-inner"
                 name="DIV_COMPOUND.inner" type="text" value="value1"/>
          <input class="required text" id="input-DIV_COMPOUND-inner2"
                 name="DIV_COMPOUND.inner2" required="required" type="text"
                 value="value2"/>
        </div>
        """, fxml(div()))

        data = div.extract({
            'DIV_COMPOUND.inner': '1',
            'DIV_COMPOUND.inner2': '2',
        })
        self.assertEqual(data.name, 'DIV_COMPOUND')
        self.assertEqual(data.value, UNSET)
        expected = odict()
        expected['inner'] = '1'
        expected['inner2'] = '2'
        self.assertEqual(data.extracted, expected)
        self.assertEqual(data.errors, [])

        data_inner = data['inner']
        self.assertEqual(data_inner.name, 'inner')
        self.assertEqual(data_inner.value, 'value1')
        self.assertEqual(data_inner.extracted, '1')
        self.assertEqual(data_inner.errors, [])

        data_inner2 = data['inner2']
        self.assertEqual(data_inner2.name, 'inner2')
        self.assertEqual(data_inner2.value, 'value2')
        self.assertEqual(data_inner2.extracted, '2')
        self.assertEqual(data_inner2.errors, [])

    def test_div_blueprint_compound_leaf(self):
        # Div blueprint as compound, but with ``leaf`` property set. Causes
        # ``hybrid_renderer`` and ``hybrid_extractor`` to skip auto delegating
        # to ``compound_renderer`` and ``compound_extractor``
        div = factory(
            'div',
            name='DIV_COMPOUND_LEAF',
            props={
                'leaf': True
            })
        div['inner'] = factory(
            'text',
            value='value1')
        div['inner2'] = factory(
            'text',
            value='value2',
            props={
                'required': True
            })
        self.assertEqual(div(), '<div></div>')

        data = div.extract({
            'DIV_COMPOUND_LEAF.inner': '1',
            'DIV_COMPOUND_LEAF.inner2': '2',
        })
        self.assertEqual(data.name, 'DIV_COMPOUND_LEAF')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, UNSET)
        self.assertEqual(data.errors, [])
        self.assertEqual(data.keys(), [])

    def test_div_blueprint_as_leaf(self):
        # Div blueprint as leaf
        input = factory(
            'div:text',
            name='DIV',
            value='1')
        self.checkOutput("""
        <div>
          <input class="text" id="input-DIV" name="DIV"
                 type="text" value="1"/>
        </div>
        """, fxml(input()))

        data = input.extract({
            'DIV': '2',
        })
        self.assertEqual(data.name, 'DIV')
        self.assertEqual(data.value, '1')
        self.assertEqual(data.extracted, '2')
        self.assertEqual(data.errors, [])

        # Empty div
        input = factory(
            'div',
            name='DIV')
        self.assertEqual(input(), '<div></div>')

        # Div with data attributes
        input = factory(
            'div',
            name='DIV',
            props={
                'data': {
                    'foo': 'bar'
                }
            })
        self.assertEqual(input(), "<div data-foo='bar'></div>")

        # Display mode
        div = factory(
            'div',
            name='DIV',
            props={
                'class': 'foo'
            },
            mode='display')
        self.assertEqual(div(), '<div class="foo"></div>')

        input = factory(
            'div:text',
            name='DIV',
            value='1',
            mode='display')
        self.checkOutput("""
        <div>
          <div class="display-text" id="display-DIV">1</div>
        </div>
        """, fxml(input()))

    def test_fieldset_blueprint(self):
        compound = factory(
            'fieldset',
            'COMPOUND',
            props={
                'legend': 'Some Test'
            })
        compound['inner'] = factory('text', 'inner', 'value')
        compound['inner2'] = factory('text', 'inner2', 'value2')
        self.checkOutput("""
        <fieldset id="fieldset-COMPOUND">
          <legend>Some Test</legend>
          <input class="text" id="input-COMPOUND-inner" name="COMPOUND.inner"
                 type="text" value="value"/>
          <input class="text" id="input-COMPOUND-inner2" name="COMPOUND.inner2"
                 type="text" value="value2"/>
        </fieldset>
        """, fxml(compound()))

        # Structural fieldset renders without id attribute
        compound = factory(
            'fieldset',
            'COMPOUND',
            props={
                'structural': True
            })
        self.assertEqual(compound(), '<fieldset></fieldset>')

        # Fieldset display renderers are the same as fieldset edit renderers
        compound = factory(
            'fieldset',
            'COMPOUND',
            props={
                'legend': 'Some Test'
            },
            mode='display')
        self.checkOutput("""
        <fieldset id="fieldset-COMPOUND">
          <legend>Some Test</legend>
        </fieldset>
        """, fxml(compound()))

    def test_form_blueprint(self):
        # Test Form
        form = factory(
            'form',
            name='FORM',
            props={
                'action': 'http://fubar.com'
            })
        self.assertEqual(form(), (
            '<form action="http://fubar.com" enctype="multipart/form-data" '
            'id="form-FORM" method="post" novalidate="novalidate"></form>'
        ))

        # Form action as callable
        def action(widget, data):
            return 'http://fubar.com'

        form = factory(
            'form',
            name='FORM',
            props={
                'action': action
            })
        self.assertEqual(form(), (
            '<form action="http://fubar.com" enctype="multipart/form-data" '
            'id="form-FORM" method="post" novalidate="novalidate"></form>'
        ))

        # Form display renderer
        form = factory(
            'form',
            name='FORM',
            props={
                'action': 'http://fubar.com'
            },
            mode='display')
        self.assertEqual(form(), '<div></div>')

        # Create a form with some children
        form = factory(
            'form',
            name='myform',
            props={
                'action': 'http://www.domain.tld/someform'
            })
        form['someinput'] = factory(
            'label:text',
            props={
                'label': 'Your Text'
            })

        self.form_data = None

        def formaction(widget, data):
            self.form_data = data

        def formnext(request):
            return 'http://www.domain.tld/result'

        form['submit'] = factory(
            'submit',
            props={
                'handler': formaction,
                'next': formnext,
                'action': True
            })

        # Render an empty form
        self.checkOutput("""
        <form action="http://www.domain.tld/someform"
              enctype="multipart/form-data" id="form-myform" method="post"
              novalidate="novalidate">
          <label for="input-myform-someinput">Your Text</label>
          <input class="text" id="input-myform-someinput"
                 name="myform.someinput" type="text" value=""/>
          <input id="input-myform-submit" name="action.myform.submit"
                 type="submit" value="submit"/>
        </form>
        """, fxml(form()))

        # Get form data out of request (request is expected dict-like)
        request = {
            'myform.someinput': 'Hello World',
            'action.myform.submit': 'submit'
        }
        Controller(form, request)

        form_data = self.form_data
        self.assertEqual(form_data.name, 'myform')
        self.assertEqual(form_data.value, UNSET)
        expected = odict()
        expected['someinput'] = 'Hello World'
        expected['submit'] = UNSET
        self.assertEqual(form_data.extracted, expected)
        self.assertEqual(form_data.errors, [])

        input_data = form_data['someinput']
        self.assertEqual(input_data.name, 'someinput')
        self.assertEqual(input_data.value, UNSET)
        self.assertEqual(input_data.extracted, 'Hello World')
        self.assertEqual(input_data.errors, [])

        # submit blueprint gets a runtime data as well, but it's never needed
        # or used so far
        submit_data = form_data['submit']
        self.assertEqual(submit_data.name, 'submit')
        self.assertEqual(submit_data.value, UNSET)
        self.assertEqual(submit_data.extracted, UNSET)
        self.assertEqual(submit_data.errors, [])

        del self.form_data

        # Form action property can be callable
        def action(widget, data):
            return 'actionfromcall'

        form = factory(
            'form',
            name='form',
            props={
                'action': action,
            })
        self.assertEqual(form(), (
            '<form action="actionfromcall" enctype="multipart/form-data" '
            'id="form-form" method="post" novalidate="novalidate"></form>'
        ))

        # Create label for field in other compound
        form = factory(
            'form',
            name='form',
            props={
                'action': 'action'
            })
        form['label'] = factory(
            'label',
            props={
                'label': 'Foo',
                'for': 'field'
            })
        form['field'] = factory('text')
        self.checkOutput("""
        <form action="action" enctype="multipart/form-data" id="form-form"
              method="post" novalidate="novalidate">
          <label for="input-form-field">Foo</label>
          <input class="text" id="input-form-field" name="form.field"
                 type="text" value=""/>
        </form>
        """, fxml(form()))



from yafowil.base import factory


factory.doc['props']['persist'] = """\
Marker that widget data should be considered in when using persistence
mechanism.
"""

factory.doc['props']['persist_target'] = """\
Target when using persistence mechanism. If not set, target is widget name.
"""

factory.doc['props']['persist_writer'] = """\
Callback used to persist data. Accepts model, target and value as parameters.
"""


def attribute_writer(model, target, value):
    """Write ``value`` to ``target`` attribute on ``model``.
    """
    setattr(model, target, value)


def write_mapping_writer(model, target, value):
    """Write ``value`` to ``target`` write mapping key on ``model``.
    """
    model[target] = value


def node_attribute_writer(model, target, value):
    """Write ``value`` to ``target`` node.attrs key on ``model``.
    """
    model.attrs[target] = value



# -*- coding: utf-8 -*-
import yafowil.utils


@yafowil.utils.entry_point(order=0)
def register():
    import yafowil.persistence
    import yafowil.common
    import yafowil.compound
    import yafowil.table  # noqa


# execute all register entry points. supposed to be used for widget and theme
# registration
for ep, cb in yafowil.utils.get_plugins('register'):
    cb()


# execute all configure entry points. supposed to be used for theme
# configuration, like setting factory defaults and defining macros.
for ep, cb in yafowil.utils.get_plugins('configure'):
    cb()  # pragma: no cover



# -*- coding: utf-8 -*-
from node.base import AttributedNode
from node.base import OrderedNode
from node.behaviors import Attributes
from node.utils import UNSET
from plumber import plumbing
from yafowil.base import factory
from yafowil.compat import BYTES_TYPE
from yafowil.compat import IS_PY2
from yafowil.compat import LONG_TYPE
from yafowil.compat import UNICODE_TYPE
from yafowil.tests import YafowilTestCase
from yafowil.utils import as_data_attrs
from yafowil.utils import attr_value
from yafowil.utils import callable_value
from yafowil.utils import convert_value_to_datatype
from yafowil.utils import convert_values_to_datatype
from yafowil.utils import cssclasses
from yafowil.utils import cssid
from yafowil.utils import data_attrs_helper
from yafowil.utils import EMPTY_VALUE
from yafowil.utils import generic_html5_attrs
from yafowil.utils import get_example
from yafowil.utils import get_example_names
from yafowil.utils import get_plugin_names
from yafowil.utils import get_plugins
from yafowil.utils import managedprops
from yafowil.utils import Tag
from yafowil.utils import tag as deprecated_tag
from yafowil.utils import vocabulary
import uuid
import yafowil.loader  # noqa


class TestUtils(YafowilTestCase):

    def test_entry_point(self):
        # Test entry_point support tools
        self.checkOutput("""
        [...(EntryPoint.parse('register = yafowil.loader:register'),
        <function register at ...)...]
        """, str(list(get_plugins())))

        self.assertEqual(list(get_plugins('nonexisting')), [])

        self.checkOutput("""
        [...'yafowil'...]
        """, str((get_plugin_names())))

        self.assertEqual(get_plugin_names('nonexisting'), [])

    def test_examples_lookup(self):
        # Test examples lookup
        self.checkOutput("""
        ['yafowil'...]
        """, str(sorted(get_example_names())))

        factory.register_macro('field', 'field:label:error', {})

        self.assertEqual(get_example('inexistent'), None)

        examples = get_example('yafowil')
        self.checkOutput("""
        Plain Text
        ----------
        ...
        """, examples[0]['doc'])

        self.assertEqual(examples[0]['title'], 'Plain Text')
        self.assertEqual(examples[0]['widget'].name, 'yafowil-plaintext')

    def test_vocabulary(self):
        # Test the Vocabulary
        self.assertEqual(vocabulary('foo'), [('foo', 'foo')])
        self.assertEqual(vocabulary({'key': 'value'}), [('key', 'value')])
        self.assertEqual(
            vocabulary(['value', 'value2']),
            [('value', 'value'), ('value2', 'value2')]
        )
        self.assertEqual(
            vocabulary([
                ('key', 'value'),
                ('key2', 'value2', 'v2.3'),
                ('key3',)
            ]),
            [('key', 'value'), ('key2', 'value2'), ('key3', 'key3')]
        )

        def callme():
            return 'bar'
        self.assertEqual(vocabulary(callme), [('bar', 'bar')])
        self.assertTrue(vocabulary(None) is None)

    def test_Tag(self):
        # Test Tag
        tag = Tag(lambda msg: msg)
        t = tag('p', b'Lorem Ipsum. ', u'Hello World!',
                class_=b'fancy', id=u'2f5b8a234ff')
        self.assertEqual(t, (
            u'<p class="fancy" id="2f5b8a234ff">Lorem Ipsum. Hello World!</p>'
        ))
        self.assertEqual(
            tag('dummy', name='foo', **{'data-foo': 'bar'}),
            u'<dummy data-foo=\'bar\' name="foo" />'
        )
        self.assertEqual(tag('dummy', name=None), u'<dummy />')
        self.assertEqual(tag('dummy', name=UNSET), u'<dummy />')
        # deprecated test
        self.assertEqual(deprecated_tag('div', 'foo'), u'<div>foo</div>')

    def test_cssid(self):
        # Test CSS id
        @plumbing(Attributes)
        class CSSTestNode(OrderedNode):
            @property
            def dottedpath(self):
                return u'.'.join([it for it in self.path if it])

        widget = CSSTestNode(name='form')

        widget.attrs['structural'] = True
        self.assertEqual(cssid(widget, 'PREFIX'), None)

        child = widget['child'] = CSSTestNode()
        self.assertEqual(cssid(child, 'PREFIX'), 'PREFIX-form-child')
        self.assertEqual(
            cssid(child, 'PREFIX', postfix='POSTFIX'),
            'PREFIX-form-child-POSTFIX'
        )

        child = widget[u'Hällo Wörld'] = CSSTestNode()
        self.assertEqual(cssid(child, 'PREFIX'), 'PREFIX-form-Hallo_World')

    def test_css_classes(self):
        # Test CSS Classes
        @plumbing(Attributes)
        class CSSTestNode(OrderedNode):
            pass

        widget = CSSTestNode()
        widget.attrs['required'] = False
        widget.attrs['required_class'] = None
        widget.attrs['required_class_default'] = 'required'
        widget.attrs['error_class'] = None
        widget.attrs['error_class_default'] = 'error'
        widget.attrs['valid_class'] = None
        widget.attrs['valid_class_default'] = 'valid'
        widget.attrs['class'] = None
        widget.attrs['class_add'] = None

        class DummyData(object):
            def __init__(self):
                self.errors = []
                self.extracted = UNSET

            @property
            def root(self):
                return self

        data = DummyData()

        self.assertEqual(cssclasses(widget, data), None)

        widget.attrs['class'] = 'foo bar'
        self.assertEqual(cssclasses(widget, data), 'bar foo')

        widget.attrs['class'] = None
        widget.attrs['required'] = True
        self.assertEqual(cssclasses(widget, data), None)

        widget.required = False
        data.errors = True
        self.assertEqual(cssclasses(widget, data), None)

        widget.attrs['error_class'] = True
        self.assertEqual(cssclasses(widget, data), 'error')

        widget.attrs['class'] = 'foo bar'
        self.assertEqual(cssclasses(widget, data), 'bar error foo')

        widget.attrs['class'] = lambda w, d: 'baz'
        self.assertEqual(cssclasses(widget, data), 'baz error')

        widget.attrs['class_add'] = lambda w, d: 'addclass_from_callable'
        self.assertEqual(
            cssclasses(widget, data),
            'addclass_from_callable baz error'
        )

        widget.attrs['class_add'] = 'addclass'
        self.assertEqual(cssclasses(widget, data), 'addclass baz error')

        widget.attrs['class'] = None
        widget.attrs['class_add'] = None
        widget.attrs['error_class'] = 'othererror'
        self.assertEqual(cssclasses(widget, data), 'othererror')

        data.errors = False
        self.assertEqual(cssclasses(widget, data), None)

        widget.attrs['required'] = True
        self.assertEqual(cssclasses(widget, data), None)

        widget.attrs['required_class'] = True
        self.assertEqual(cssclasses(widget, data), 'required')

        widget.attrs['required_class'] = 'otherrequired'
        self.assertEqual(cssclasses(widget, data), 'otherrequired')

        widget.attrs['error_class'] = True
        data.errors = True
        widget.attrs['required_class'] = 'required'
        self.assertEqual(cssclasses(widget, data), 'error required')

        widget.attrs['class'] = 'foo bar'
        self.assertEqual(
            cssclasses(widget, data),
            'bar error foo required'
        )

        self.assertEqual(
            cssclasses(widget, data, additional=['zika', 'akiz']),
            'akiz bar error foo required zika'
        )

        data.errors = []
        data.extracted = True
        widget.attrs['required'] = False
        widget.attrs['class'] = None

        widget.attrs['valid_class'] = True
        self.assertEqual(cssclasses(widget, data), 'valid')

        widget.attrs['valid_class'] = 'custom_valid'
        self.assertEqual(cssclasses(widget, data), 'custom_valid')

    def test_managedprops(self):
        # Test managedprops annotation
        @managedprops('foo', 'bar')
        def somefunc(a, b, c):
            return a, b, c

        self.assertEqual(somefunc(1, 2, 3), (1, 2, 3))

        self.assertEqual(
            somefunc.__yafowil_managed_props__,
            ('foo', 'bar')
        )

    def test_attr_value(self):
        # Test attr_value
        widget = AttributedNode()
        data = AttributedNode()

        widget.attrs['attr'] = 'value'
        self.assertEqual(attr_value('attr', widget, data), 'value')

        def func_callback(widget, data):
            return 'func_callback value'

        widget.attrs['attr'] = func_callback
        self.assertEqual(
            attr_value('attr', widget, data),
            'func_callback value'
        )

        def failing_func_callback(widget, data):
            raise Exception('failing_func_callback')

        widget.attrs['attr'] = failing_func_callback
        with self.assertRaises(Exception) as arc:
            attr_value('attr', widget, data)
        self.assertEqual(str(arc.exception), 'failing_func_callback')

        class FormContext(object):
            def instance_callback(self, widget, data):
                return 'instance_callback'

            def failing_instance_callback(self, widget, data):
                raise Exception('failing_instance_callback')

        context = FormContext()
        widget.attrs['attr'] = context.instance_callback
        self.assertEqual(
            attr_value('attr', widget, data),
            'instance_callback'
        )

        widget.attrs['attr'] = context.failing_instance_callback
        with self.assertRaises(Exception) as arc:
            attr_value('attr', widget, data)
        self.assertEqual(str(arc.exception), 'failing_instance_callback')

    def test_as_data_attrs(self):
        self.assertTrue(as_data_attrs is generic_html5_attrs)
        html5_attrs = as_data_attrs({
            'foo': 'bar',
            'baz': ['bam'],
            'nada': None,
            'unset': UNSET
        })
        self.assertEqual(
            html5_attrs,
            {'data-baz': '["bam"]', 'data-foo': 'bar'}
        )
        html5_attrs = as_data_attrs(None)
        self.assertEqual(html5_attrs, {})
        html5_attrs = as_data_attrs(UNSET)
        self.assertEqual(html5_attrs, {})

    def test_data_attrs_helper(self):
        data = AttributedNode()
        widget = AttributedNode()
        widget.attrs['testattr1'] = 'value'
        widget.attrs['testattr2'] = True
        widget.attrs['testattr3'] = False
        widget.attrs['testattr4'] = None
        widget.attrs['testattr5'] = ['item1', 'item2', 'item3']
        widget.attrs['testattr6'] = {
            'key1': 'item1',
            'key2': 'item2',
            'key3': 'item3'
        }
        widget.attrs['testattr7'] = 1234
        widget.attrs['testattr8'] = 1234.5678
        widget.attrs['testattr9'] = UNSET
        widget.attrs['camelAttrName'] = 'camelValue'
        data_attrs_keys = [
            'testattr1', 'testattr2', 'testattr3', 'testattr4', 'testattr5',
            'testattr6', 'testattr7', 'testattr8', 'camelAttrName'
        ]
        data_attrs = data_attrs_helper(widget, data, data_attrs_keys)

        self.assertEqual(data_attrs['data-testattr1'], 'value')
        self.assertEqual(data_attrs['data-testattr2'], 'true')
        self.assertEqual(data_attrs['data-testattr3'], 'false')
        self.assertFalse('data-testattr4' in data_attrs)
        self.assertEqual(
            data_attrs['data-testattr5'],
            '["item1", "item2", "item3"]'
        )
        self.assertTrue(data_attrs['data-testattr6'].find('"key1": "item1"'))
        self.assertTrue(data_attrs['data-testattr6'].find('"key2": "item2"'))
        self.assertTrue(data_attrs['data-testattr6'].find('"key3": "item3"'))
        self.assertTrue(data_attrs['data-testattr6'].startswith('{'))
        self.assertTrue(data_attrs['data-testattr6'].endswith('}'))
        self.assertEqual(data_attrs['data-testattr7'], '1234')
        self.assertEqual(data_attrs['data-testattr8'], '1234.5678')
        self.assertFalse('data-testattr9' in data_attrs)
        self.assertEqual(data_attrs['data-camel-attr-name'], 'camelValue')

        # Test with Tag renderer
        tag = Tag(lambda msg: msg)
        self.checkOutput("""
        <dummy
          data-camel-attr-name='camelValue'
          data-testattr1='value'
          data-testattr2='true'
          data-testattr3='false'
          data-testattr5='["item1", "item2", "item3"]'
          data-testattr6='{"..."}'
          data-testattr7='1234'
          data-testattr8='1234.5678'
          name="foo" />
        """, tag('dummy', name='foo', **data_attrs))

    def test_EMPTY_VALUE(self):
        self.assertEqual(repr(EMPTY_VALUE), '<EMPTY_VALUE>')
        self.assertEqual(str(EMPTY_VALUE), '')
        self.assertFalse(bool(EMPTY_VALUE))
        self.assertEqual(len(EMPTY_VALUE), 0)

    def test_convert_value_to_datatype(self):
        # Unknown string identifier
        with self.assertRaises(KeyError) as arc:
            convert_value_to_datatype('val', 'inexistent')
        self.assertEqual(str(arc.exception), "'inexistent'")

        # Function returns ``EMPTY_VALUE`` marker if value is ``None`` or empty
        # string
        self.assertEqual(convert_value_to_datatype('', 'uuid'), EMPTY_VALUE)
        self.assertEqual(convert_value_to_datatype(None, 'uuid'), EMPTY_VALUE)

    def test_convert_value_to_datatype_str(self):
        # Convert to string by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'str'), UNSET)

        converted = convert_value_to_datatype(u'string', 'str')
        self.assertEqual(converted, b'string')
        self.assertTrue(isinstance(converted, BYTES_TYPE))

        with self.assertRaises(UnicodeEncodeError) as arc:
            convert_value_to_datatype(u'äöü', 'str')
        self.checkOutput("""
        'ascii' codec can't encode character...: ordinal not in range(128)
        """, str(arc.exception))

        # Convert to string by type
        self.assertEqual(convert_value_to_datatype(UNSET, BYTES_TYPE), UNSET)

        converted = convert_value_to_datatype(u'string', BYTES_TYPE)
        self.assertEqual(converted, b'string')
        self.assertTrue(isinstance(converted, BYTES_TYPE))

        with self.assertRaises(UnicodeEncodeError) as arc:
            convert_value_to_datatype(u'äöü', BYTES_TYPE)
        self.checkOutput("""
        'ascii' codec can't encode character...: ordinal not in range(128)
        """, str(arc.exception))

    def test_convert_value_to_datatype_unicode(self):
        # Convert to unicode by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'unicode'), UNSET)

        converted = convert_value_to_datatype('unicode', 'unicode')
        self.assertEqual(converted, u'unicode')
        self.assertTrue(isinstance(converted, UNICODE_TYPE))

        with self.assertRaises(UnicodeDecodeError) as arc:
            convert_value_to_datatype(b'\xc3\xa4\xc3\xb6\xc3\xbc', 'unicode')
        msg = (
            "'ascii' codec can't decode byte 0xc3 in position 0: "
            "ordinal not in range(128)")
        self.assertEqual(str(arc.exception), msg)

        # Convert to unicode by type
        self.assertEqual(convert_value_to_datatype(UNSET, UNICODE_TYPE), UNSET)

        converted = convert_value_to_datatype('unicode', UNICODE_TYPE)
        self.assertEqual(converted, u'unicode')
        self.assertTrue(isinstance(converted, UNICODE_TYPE))

        with self.assertRaises(UnicodeDecodeError) as arc:
            convert_value_to_datatype(b'\xc3\xa4\xc3\xb6\xc3\xbc', UNICODE_TYPE)
        msg = (
            "'ascii' codec can't decode byte 0xc3 in position 0: "
            "ordinal not in range(128)")
        self.assertEqual(str(arc.exception), msg)

    def test_convert_value_to_datatype_int(self):
        # Convert to int by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'int'), UNSET)

        converted = convert_value_to_datatype('1', 'int')
        self.assertEqual(converted, 1)
        self.assertTrue(isinstance(converted, int))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('1.0', 'int')
        msg = "invalid literal for int() with base 10: '1.0'"
        self.assertEqual(str(arc.exception), msg)

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('a', 'int')
        msg = "invalid literal for int() with base 10: 'a'"
        self.assertEqual(str(arc.exception), msg)

        converted = convert_value_to_datatype(2.0, 'int')
        self.assertEqual(converted, 2)
        self.assertTrue(isinstance(converted, int))

        # Convert to int by type
        self.assertEqual(convert_value_to_datatype(UNSET, int), UNSET)

        converted = convert_value_to_datatype('3', int)
        self.assertEqual(converted, 3)
        self.assertTrue(isinstance(converted, int))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('2.0', int)
        msg = "invalid literal for int() with base 10: '2.0'"
        self.assertEqual(str(arc.exception), msg)

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', int)
        msg = "invalid literal for int() with base 10: 'b'"
        self.assertEqual(str(arc.exception), msg)

        converted = convert_value_to_datatype(4.0, int)
        self.assertEqual(converted, 4)
        self.assertTrue(isinstance(converted, int))

    def test_convert_value_to_datatype_long(self):
        # Convert to long by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'long'), UNSET)

        converted = convert_value_to_datatype('1', 'long')
        self.assertEqual(converted, LONG_TYPE(1))
        self.assertTrue(isinstance(converted, LONG_TYPE))

        converted = convert_value_to_datatype(2.0, 'long')
        self.assertEqual(converted, LONG_TYPE(2))
        self.assertTrue(isinstance(converted, LONG_TYPE))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('a', 'long')
        # there is no long type in python 3, falls back to int
        msg = (
            "invalid literal for long() with base 10: 'a'"
            if IS_PY2
            else "invalid literal for int() with base 10: 'a'"
        )
        self.assertEqual(str(arc.exception), msg)

        # Convert to long by type
        self.assertEqual(convert_value_to_datatype(UNSET, LONG_TYPE), UNSET)

        converted = convert_value_to_datatype('3', LONG_TYPE)
        self.assertEqual(converted, LONG_TYPE(3))
        self.assertTrue(isinstance(converted, LONG_TYPE))

        converted = convert_value_to_datatype(4.0, LONG_TYPE)
        self.assertEqual(converted, LONG_TYPE(4))
        self.assertTrue(isinstance(converted, LONG_TYPE))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', LONG_TYPE)
        # there is no long type in python 3, falls back to int
        msg = (
            "invalid literal for long() with base 10: 'b'"
            if IS_PY2
            else "invalid literal for int() with base 10: 'b'"
        )
        self.assertEqual(str(arc.exception), msg)

    def test_convert_value_to_datatype_float(self):
        # Convert to float by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'float'), UNSET)

        converted = convert_value_to_datatype('1,0', 'float')
        self.assertEqual(converted, 1.0)
        self.assertTrue(isinstance(converted, float))

        converted = convert_value_to_datatype('2', 'float')
        self.assertEqual(converted, 2.0)
        self.assertTrue(isinstance(converted, float))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('a', 'float')
        msg = (
            "could not convert string to float: a"
            if IS_PY2
            else "could not convert string to float: 'a'"
        )
        self.assertEqual(str(arc.exception), msg)

        converted = convert_value_to_datatype(3, 'float')
        self.assertEqual(converted, 3.0)
        self.assertTrue(isinstance(converted, float))

        # Convert to float by type
        self.assertEqual(convert_value_to_datatype(UNSET, float), UNSET)

        converted = convert_value_to_datatype('4,0', float)
        self.assertEqual(converted, 4.0)
        self.assertTrue(isinstance(converted, float))

        converted = convert_value_to_datatype('5', float)
        self.assertEqual(converted, 5.0)
        self.assertTrue(isinstance(converted, float))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', float)
        msg = (
            "could not convert string to float: b"
            if IS_PY2
            else "could not convert string to float: 'b'"
        )
        self.assertEqual(str(arc.exception), msg)

        converted = convert_value_to_datatype(6, float)
        self.assertEqual(converted, 6.0)
        self.assertTrue(isinstance(converted, float))

    def test_convert_value_to_datatype_uuid(self):
        # Convert to uuid by id
        self.assertEqual(convert_value_to_datatype(UNSET, 'uuid'), UNSET)

        converted = convert_value_to_datatype(str(uuid.uuid4()), 'uuid')
        self.assertTrue(isinstance(converted, uuid.UUID))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('a', 'uuid')
        msg = 'badly formed hexadecimal UUID string'
        self.assertEqual(str(arc.exception), msg)

        # Convert to uuid by type
        self.assertEqual(convert_value_to_datatype(UNSET, uuid.UUID), UNSET)

        converted = convert_value_to_datatype(str(uuid.uuid4()), uuid.UUID)
        self.assertTrue(isinstance(converted, uuid.UUID))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('a', uuid.UUID)
        msg = 'badly formed hexadecimal UUID string'
        self.assertEqual(str(arc.exception), msg)

    def test_convert_value_to_datatype_function(self):
        # Custom converter as function
        def convert_func(val):
            if val == 'a':
                return 'convertet: {0}'.format(val)
            raise ValueError("Value not 'a'")

        self.assertEqual(
            convert_value_to_datatype('a', convert_func),
            'convertet: a'
        )

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', convert_func)
        self.assertEqual(str(arc.exception), "Value not 'a'")

    def test_convert_value_to_datatype_class(self):
        # Custom converters as class
        class Converter(object):

            def __init__(self, val):
                if val != 'a':
                    raise ValueError("Value not 'a'")

        converted = convert_value_to_datatype('a', Converter)
        self.assertTrue(isinstance(converted, Converter))

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', Converter)
        self.assertEqual(str(arc.exception), "Value not 'a'")

    def test_convert_value_to_datatype_instance(self):
        # Custom converter as class instance with ``__call__`` function
        class ConverterInst(object):

            def __call__(self, val):
                if val != 'a':
                    raise ValueError("Value not 'a'")
                return 'convertet: {0}'.format(val)

        self.assertEqual(
            convert_value_to_datatype('a', ConverterInst()),
            'convertet: a'
        )

        with self.assertRaises(ValueError) as arc:
            convert_value_to_datatype('b', ConverterInst())
        self.assertEqual(str(arc.exception), "Value not 'a'")

    def test_convert_values_to_datatype(self):
        self.assertEqual(convert_values_to_datatype(UNSET, 'int'), UNSET)
        self.assertEqual(convert_values_to_datatype([UNSET], 'int'), [UNSET])
        self.assertEqual(convert_values_to_datatype('0', int), 0)
        self.assertEqual(convert_values_to_datatype(['0', '1'], int), [0, 1])

    def test_callable_value(self):
        # non callbale is returned as is
        self.assertEqual(callable_value('1', None, None), '1')

        # callable expects widget and data
        def dummy_callable(widget, data):
            return "2"
        self.assertEqual(callable_value(dummy_callable, None, None), '2')

        # callable with no parameters raises type error
        def invalid_signature():
            pass  # pragma: no cover
        self.assertRaises(TypeError, invalid_signature, None, None)



from yafowil.base import factory
from yafowil.compound import compound_extractor
from yafowil.compound import compound_renderer
from yafowil.compound import hybrid_extractor
from yafowil.utils import attr_value
from yafowil.utils import css_managed_props
from yafowil.utils import cssclasses
from yafowil.utils import managedprops


###############################################################################
# table
###############################################################################

@managedprops('id', *css_managed_props)
def table_renderer(widget, data):
    attrs = {
        'id': attr_value('id', widget, data),
        'class_': cssclasses(widget, data),
    }
    return data.tag('table', data.rendered, **attrs)


factory.register(
    'table',
    extractors=[compound_extractor],
    edit_renderers=[
        compound_renderer,
        table_renderer
    ],
    display_renderers=[
        compound_renderer,
        table_renderer
    ])

factory.doc['blueprint']['table'] = """\
``<table>`` compound widget for table creation.
"""

factory.doc['props']['table.id'] = """\
Value of table id attribute.
"""


###############################################################################
# thead
###############################################################################

def thead_renderer(widget, data):
    return data.tag('thead', data.rendered)


factory.register(
    'thead',
    extractors=[compound_extractor],
    edit_renderers=[
        compound_renderer,
        thead_renderer
    ],
    display_renderers=[
        compound_renderer,
        thead_renderer
    ])

factory.doc['blueprint']['thead'] = """\
``<thead>`` compound widget for table creation.
"""


###############################################################################
# tbody
###############################################################################

def tbody_renderer(widget, data):
    return data.tag('tbody', data.rendered)


factory.register(
    'tbody',
    extractors=[compound_extractor],
    edit_renderers=[
        compound_renderer,
        tbody_renderer
    ],
    display_renderers=[
        compound_renderer,
        tbody_renderer
    ])

factory.doc['blueprint']['tbody'] = """\
``<tbody>`` compound widget for table creation.
"""


###############################################################################
# tr
###############################################################################

@managedprops('id', *css_managed_props)
def tr_renderer(widget, data):
    attrs = {
        'id': attr_value('id', widget, data),
        'class_': cssclasses(widget, data),
    }
    return data.tag('tr', data.rendered, **attrs)


factory.register(
    'tr',
    extractors=[compound_extractor],
    edit_renderers=[
        compound_renderer,
        tr_renderer
    ],
    display_renderers=[
        compound_renderer,
        tr_renderer
    ])

factory.doc['blueprint']['tr'] = """\
``<tr>`` compound widget for table creation.
"""

factory.doc['props']['tr.id'] = """\
Value of id attribute.
"""


###############################################################################
# th
###############################################################################

@managedprops('id', 'rowspan', 'colspan', 'label', *css_managed_props)
def th_renderer(widget, data):
    attrs = {
        'id': attr_value('id', widget, data),
        'class_': cssclasses(widget, data),
        'colspan': attr_value('colspan', widget, data),
        'rowspan': attr_value('rowspan', widget, data),
    }
    contents = attr_value('label', widget, data)
    if not contents:
        contents = data.rendered
    return data.tag('th', contents, **attrs)


factory.register(
    'th',
    edit_renderers=[th_renderer],
    display_renderers=[th_renderer])

factory.doc['blueprint']['th'] = """\
``<th>`` compound widget for table creation.
"""

factory.doc['props']['th.id'] = """\
Value of id attribute.
"""

factory.doc['props']['th.rowspan'] = """\
Value of rowspan attribute.
"""

factory.doc['props']['th.colspan'] = """\
Value of colspan attribute.
"""

factory.doc['props']['th.label'] = """\
Explicit th content. If absent, rendered markup from downstream blueprint(s)
is used.
"""


###############################################################################
# td
###############################################################################

@managedprops('id', 'rowspan', 'colspan', *css_managed_props)
def td_renderer(widget, data):
    attrs = {
        'id': attr_value('id', widget, data),
        'class_': cssclasses(widget, data),
        'colspan': attr_value('colspan', widget, data),
        'rowspan': attr_value('rowspan', widget, data),
    }
    if len(widget):
        rendered = compound_renderer(widget, data)
    else:
        rendered = data.rendered
    return data.tag('td', rendered, **attrs)


factory.register(
    'td',
    extractors=[hybrid_extractor],
    edit_renderers=[td_renderer],
    display_renderers=[td_renderer])

factory.doc['blueprint']['td'] = """\
``<td>`` compound widget for table creation.
"""

factory.doc['props']['td.id'] = """\
Value of id attribute.
"""

factory.doc['props']['td.rowspan'] = """\
Value of rowspan attribute.
"""

factory.doc['props']['td.colspan'] = """\
Value of colspan attribute.
"""



from node.tests import NodeTestCase
from node.tests import patch
from yafowil import resources
from yafowil.resources import YafowilResources


class MockFactory(object):
    mock_resources = {
        'yafowil.widget.foo': {
            'resourcedir': '/path/to/yafowil.widget.foo/resources',
            'css': [{
                'group': 'yafowil.widget.foo.common',
                'resource': 'widget.css',
                'order': 10
            }],
            'js': [{
                'group': 'yafowil.widget.foo.common',
                'resource': 'widget.js',
                'order': 10
            }]
        },
        'yafowil.widget.bar': {
            'resourcedir': '/path/to/yafowil.widget.bar/resources',
            'css': [{
                'group': 'yafowil.widget.bar.common',
                'resource': 'widget.css',
                'order': 10
            }],
            'js': [{
                'group': 'yafowil.widget.bar.common',
                'resource': 'widget.js',
                'order': 10
            }]
        },
    }

    def resources_for(self, plugin_name):
        return self.mock_resources.get(plugin_name)


def mock_get_plugin_names():
    return [
        'yafowil.widget.foo',
        'yafowil.widget.bar',
        'yafowil.widget.baz',
    ]


class TestResources(NodeTestCase):

    @patch(resources, 'factory', MockFactory())
    @patch(resources, 'get_plugin_names', mock_get_plugin_names)
    def test_resources(self):
        resources = YafowilResources()
        self.assertEqual(resources.js_resources, [
            'yafowil.widget.foo/widget.js',
            'yafowil.widget.bar/widget.js'
        ])
        self.assertEqual(resources.css_resources, [
            'yafowil.widget.foo/widget.css',
            'yafowil.widget.bar/widget.css'
        ])
        resources = YafowilResources(
            js_skip='yafowil.widget.foo.common',
            css_skip='yafowil.widget.bar.common'
        )
        self.assertEqual(resources.js_resources, [
            'yafowil.widget.bar/widget.js'
        ])
        self.assertEqual(resources.css_resources, [
            'yafowil.widget.foo/widget.css'
        ])



from __future__ import print_function
from node.tests import NodeTestCase
from yafowil.base import factory
from yafowil.compat import IS_PY2
import lxml.etree as etree
import sys
import unittest
import yafowil.common
import yafowil.compound
import yafowil.persistence
import yafowil.table


if not IS_PY2:
    from importlib import reload


class YafowilTestCase(NodeTestCase):

    def setUp(self):
        super(YafowilTestCase, self).setUp()
        factory.push_state()
        factory.clear()
        reload(yafowil.persistence)
        reload(yafowil.common)
        reload(yafowil.compound)
        reload(yafowil.table)

    def tearDown(self):
        factory.pop_state()
        super(YafowilTestCase, self).tearDown()


def fxml(xml):
    et = etree.fromstring(xml)
    return etree.tostring(et, pretty_print=True).decode('utf-8')


def test_suite():
    from yafowil.tests import test_base
    from yafowil.tests import test_common
    from yafowil.tests import test_compound
    from yafowil.tests import test_controller
    from yafowil.tests import test_persistence
    from yafowil.tests import test_resources
    from yafowil.tests import test_table
    from yafowil.tests import test_tsf
    from yafowil.tests import test_utils

    suite = unittest.TestSuite()

    suite.addTest(unittest.findTestCases(test_base))
    suite.addTest(unittest.findTestCases(test_common))
    suite.addTest(unittest.findTestCases(test_compound))
    suite.addTest(unittest.findTestCases(test_controller))
    suite.addTest(unittest.findTestCases(test_persistence))
    suite.addTest(unittest.findTestCases(test_resources))
    suite.addTest(unittest.findTestCases(test_table))
    suite.addTest(unittest.findTestCases(test_tsf))
    suite.addTest(unittest.findTestCases(test_utils))

    return suite


def run_tests():
    from zope.testrunner.runner import Runner

    runner = Runner(found_suites=[test_suite()])
    runner.run()
    sys.exit(int(runner.failed))


if __name__ == '__main__':
    run_tests()



from setuptools import find_packages
from setuptools import setup
from setuptools.command.test import test
import os


def read_file(name):
    with open(os.path.join(os.path.dirname(__file__), name)) as f:
        return f.read()


version = '3.0'
shortdesc = 'YAFOWIL - declarative, framework independent, flexible HTML forms'
longdesc = '\n\n'.join([read_file(name) for name in [
    'README.rst',
    'CHANGES.rst',
    'LICENSE.rst'
]])


class Test(test):

    def run_tests(self):
        from yafowil import tests
        tests.run_tests()


setup(
    name='yafowil',
    version=version,
    description=shortdesc,
    long_description=longdesc,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ],
    keywords='html form rendering processing',
    author='Yafowil Contributors',
    author_email='dev@conestack.org',
    url=u'http://github.com/conestack/yafowil',
    license='Simplified BSD',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['yafowil'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'node>=1.0',
        'setuptools',
        'webresource'
    ],
    tests_require=[
        'lxml',
        'zope.testrunner'
    ],
    test_suite='yafowil.tests.test_suite',
    extras_require=dict(test=[
        'lxml',
        'zope.testrunner'
    ]),
    entry_points="""
    [yafowil.plugin]
    register = yafowil.loader:register
    example = yafowil.example:get_example
    """
)



# -*- coding: utf-8 -*-
from yafowil.base import factory
from yafowil.utils import get_plugin_names
import copy


class YafowilResources(object):
    """Object which can be used as base for resource publishing specific
    framework integration code.

    This class belongs to the B/C theme registration and is deprecated.
    It will be removed in future versions.
    """

    def __init__(self, js_skip=[], css_skip=[]):
        """Initialize YAFOWIL resources.

        This object normally gets instanciated only once at application
        startup.

        :param js_skip: ignored resource groups when aggregating JS resources.
        :param css_skip: ignored resource groups when aggregating CSS
                         resources.
        """
        all_js = list()
        all_css = list()
        for plugin_name in get_plugin_names():
            resources = factory.resources_for(plugin_name)
            if not resources:
                continue
            resources = copy.deepcopy(resources)
            resource_base = self.configure_resource_directory(
                plugin_name, resources['resourcedir'])
            for js in resources['js']:
                if js['group'] in js_skip:
                    continue
                if not self._is_remote_resource(js['resource']):
                    js['resource'] = resource_base + '/' + js['resource']
                all_js.append(js)
            for css in resources['css']:
                if css['group'] in css_skip:
                    continue
                if not self._is_remote_resource(css['resource']):
                    css['resource'] = resource_base + '/' + css['resource']
                all_css.append(css)
        all_js = sorted(all_js, key=lambda x: x['order'])
        all_css = sorted(all_css, key=lambda x: x['order'])
        self.js_resources = [res['resource'] for res in all_js]
        self.css_resources = [res['resource'] for res in all_css]

    def _is_remote_resource(self, resource):
        return resource.startswith('http://') \
            or resource.startswith('https://') \
            or resource.startswith('//')

    def configure_resource_directory(self, plugin_name, resourc_edir):
        """Register plugin specific resource directory to web publisher and
        return the base URL under which the plugin specific resources are
        available.

        This function is supposed to be implemented on derived object by
        framework integration code.

        :param plugin_name: plugin name.
        :param resource_dir: absolute path of the physical location of plugin
            resource directory.
        :return string: base URL under which the plugin specific resources are
            available.
        """
        return plugin_name



from yafowil.base import factory
from yafowil.controller import Controller
from yafowil.tests import fxml
from yafowil.tests import YafowilTestCase
import yafowil.common
import yafowil.compound  # noqa


class TestController(YafowilTestCase):

    def test_controller(self):
        # Dummy context
        class Context(object):
            value = 'hello world'
        context = Context()

        # Dummy getter
        def getter(widget, data):
            return data.request.context.value

        # Create Widget tree
        form = factory(
            u'form',
            name='testform',
            props={
                'action': 'http://fubar.com'
            })
        form['field1'] = factory(
            'text',
            value=getter)
        form['field2'] = factory(
            'text',
            value='',
            props={
                'required': True
            })

        # Define action ``handler``
        def handler(widget, data):
            self.handler_result = 'handler called "%s"' % '.'.join(widget.path)

        # Define action ``next``
        def next(request):
            return 'next return value'

        # Indicate widget to be an ``action`` definition by setting ``action``
        # attribute to widget properties. ``expression``, ``handler`` and
        # ``next`` are action referring properties
        props = {
            'action': 'save',
            'expression': True,
            'handler': handler,
            'next': next,
            'label': 'Save',
            'skip': False,
        }

        # Add save action
        form['save'] = factory('submit', props=props)

        # Add cancel action. In this case we want the form processing to be
        # skipped and just the next action to be performed
        props = {
            'action': 'cancel',
            'expression': True,
            'handler': None,
            'next': next,
            'label': 'Cancel',
            'skip': True,
        }
        form['cancel'] = factory('submit', props=props)

        # Check widget tree
        self.assertEqual(form.treerepr().split('\n'), [
            "<class 'yafowil.base.Widget'>: testform",
            "  <class 'yafowil.base.Widget'>: field1",
            "  <class 'yafowil.base.Widget'>: field2",
            "  <class 'yafowil.base.Widget'>: save",
            "  <class 'yafowil.base.Widget'>: cancel",
            ""
        ])

        # Dummy request
        class Request(dict):
            context = None

        Request.context = context
        request = Request()

        # Render form with empty request
        data = form.extract(request)
        self.checkOutput("""
        <form action="http://fubar.com" enctype="multipart/form-data"
              id="form-testform" method="post" novalidate="novalidate">
          <input class="text" id="input-testform-field1" name="testform.field1"
                 type="text" value="hello world"/>
          <input class="required text" id="input-testform-field2"
                 name="testform.field2" required="required" type="text"
                 value=""/>
          <input id="input-testform-save" name="action.testform.save"
                 type="submit" value="Save"/>
          <input id="input-testform-cancel" name="action.testform.cancel"
                 type="submit" value="Cancel"/>
        </form>
        """, fxml(form(data)))

        # Create controller for form
        controller = Controller(form, request)

        # If action is not triggered, or ``action['next']`` is not set,
        # ``Controller.next`` is ``None``
        self.assertEqual(controller.next, None)

        # An empty request does not trigger validation failures
        self.assertFalse(controller.error)

        # Provide empty required field and it fails!
        request['testform.field2'] = ''
        controller = Controller(form, request)
        self.assertTrue(controller.error)

        # Provide required field and all is fine
        request['testform.field2'] = '1'
        controller = Controller(form, request)
        self.assertFalse(controller.error)

        # Trigger save action without required field
        request['testform.field2'] = ''
        request['action.testform.save'] = '1'
        controller = Controller(form, request)
        self.assertTrue(controller.error, True)
        self.assertTrue(controller.performed, True)

        # Trigger save action with valid input
        request['testform.field2'] = '1'
        controller = Controller(form, request)
        self.assertEqual(self.handler_result, 'handler called "testform"')
        self.handler_result = None
        self.assertEqual(controller.next, 'next return value')
        self.assertFalse(controller.error)
        self.assertTrue(controller.performed)

        # Render the form performed
        self.checkOutput("""
        <form action="http://fubar.com" enctype="multipart/form-data"
              id="form-testform" method="post" novalidate="novalidate">
          <input class="text" id="input-testform-field1" name="testform.field1"
                 type="text" value="hello world"/>
          <input class="required text" id="input-testform-field2"
                 name="testform.field2" required="required" type="text"
                 value="1"/>
          <input id="input-testform-save" name="action.testform.save"
                 type="submit" value="Save"/>
          <input id="input-testform-cancel" name="action.testform.cancel"
                 type="submit" value="Cancel"/>
        </form>
        """, fxml(controller.rendered))

        # Trigger cancel action. performing is skipped
        del request['action.testform.save']
        request['action.testform.cancel'] = '1'
        controller = Controller(form, request)
        self.assertEqual(controller.next, 'next return value')
        self.assertFalse(controller.performed)

        # Render form not performed
        self.checkOutput("""
        <form action="http://fubar.com" enctype="multipart/form-data"
              id="form-testform" method="post" novalidate="novalidate">
          <input class="text" id="input-testform-field1" name="testform.field1"
                 type="text" value="hello world"/>
          <input class="required text" id="input-testform-field2"
                 name="testform.field2" required="required" type="text"
                 value=""/>
          <input id="input-testform-save" name="action.testform.save"
                 type="submit" value="Save"/>
          <input id="input-testform-cancel" name="action.testform.cancel"
                 type="submit" value="Cancel"/>
        </form>
        """, fxml(controller.rendered))

        # Try recursive lookup of actions
        form = factory(
            u'form',
            name='testform',
            props={
                'action': 'http://fubar.com'
            })
        form['level1'] = factory(
            'submit',
            props={
                'action': 'l1action'
            })
        form['fieldset'] = factory('fieldset')
        form['fieldset']['level2'] = factory(
            'submit',
            props={
                'action': 'l2action'
            })
        form['fieldset']['subset'] = factory('fieldset')
        form['fieldset']['subset']['level3'] = factory(
            'submit',
            props={
                'action': 'l3action'
            })
        controller = Controller(form, {})
        self.assertEqual(len(controller.actions), 3)
        self.assertEqual(controller.actions[0].name, 'level1')
        self.assertEqual(controller.actions[1].name, 'level2')
        self.assertEqual(controller.actions[2].name, 'level3')
